from cli import pyke, commands
from .version_cmd import version
import click

@click.group(help='This CLI will allow you to interact with the Lumavate platform from your terminal. For setup instructions, look in github. \
Each command below has subcommands. Pass the --help flag to those commands for more information on how to use them.')
def cli():
    pass

cli.add_command(commands.env)
cli.add_command(commands.profile)
cli.add_command(commands.organization)
cli.add_command(commands.widget)
cli.add_command(commands.widget_version)
cli.add_command(commands.component_set)
cli.add_command(commands.component_set_version)
cli.add_command(commands.microservice)
cli.add_command(commands.microservice_version)
cli.add_command(version)
cli.add_command(commands.api)

if __name__ == '__main__':
    cli(help='')



from cli import pyke, behavior
import click

@click.group('component-set', help='A component-set is comprised of component-set-versions. You can share component-sets with other orgs giving them access to use the associated versions.')
def component_set():
    pass

@click.command('ls')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def list(format, profile, json, **kwargs):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json, **kwargs).list()

@click.command('rm')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to delete')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def delete(component_set, profile, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).delete(component_set)

@click.command('add')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--name', prompt=True, help='The name of the component set')
@click.option('--url-ref', prompt=True, help="A unique value for how the component-set will be addressed in a URL.")
@click.option('--icon-file', '-path', type=click.Path(), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def create(profile, name, url_ref, icon_file, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).create(name, url_ref, icon_file)

@click.command()
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to update')
@click.option('--name', default='', help='The name of the component set')
@click.option('--icon-file', '-path', default='', type=click.Path(), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def update(component_set, name, icon_file, profile, format, json):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).update(component_set, name, icon_file)

@click.command('access', help="The access command shares and unshares objects with child orgs.")
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set being shared/unshared')
@click.option('--add', multiple=True, help="Share with an org. (ID || Org Name)")
@click.option('--rm', multiple=True, help="Un-Share with an org. (ID || Org Name)")
@click.option('--absolute', multiple=True, help='Flush and fill shared with list. (ID || Org Name)')
@click.option('--format', '-f', default='{failed} {sharedWith} {unsharedFrom} {resultingGrantees}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@click.option('--current', is_flag=True, help="Returns a list of orgs currently shared with this object.")
@pyke.auth.profile_context_required
def share(component_set, add, rm, absolute, profile, format, json, current):
    behavior.ComponentSetBehavior(profile=profile, format=format, json=json).share(component_set, add, rm, absolute, current)

component_set.add_command(list)
component_set.add_command(create)
component_set.add_command(delete)
component_set.add_command(update)
component_set.add_command(share)



from pathlib import Path
import docker
import shutil
import click
import gzip
import time
import sys
import os

class ContainerUtil:
  def __init__(self):
    try:
      self.client = docker.Client(base_url='unix://var/run/docker.sock')
    except Exception as e:
      raise click.ClickException(click.style("Error communicating with the docker daemon. Are you sure it's running? {}".format(e), fg='red'))

  def save_image(self, image_name):
    image = self.client.get_image(image_name)
    file_path = '{}/{}'.format(str(Path.home()), image_name)

    with open(file_path, 'w+b') as image_tar:
      image_tar.write(image.data)

    zipped_file_path = self.zip_image(file_path)

    # Cleanup
    os.remove(file_path)

    return zipped_file_path

  def zip_image(self, file_path):
    with open(file_path, 'rb') as f_in:
      with gzip.open('{}.tar.gz'.format(file_path), 'w+b') as f_out:
        shutil.copyfileobj(f_in, f_out)

    return '{}.tar.gz'.format(file_path)



from ast import literal_eval
from pathlib import Path
from cli import pyke
import requests
import click
import time
import json
import sys
import os

class MicroserviceVersionBehavior:
    def __init__(self, profile=None, microservice=None, format=None, filter=None, page=None, pagesize=None, json=False):
        self.util = pyke.Util(profile=profile)
        self.object_type = 'microservice'
        self.json = json
        self.profile = profile
        self.microservice = microservice
        self.format = format
        self.filter = filter
        self.page = page
        self.pagesize = pagesize

    def list(self, microservice):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)
        data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

        handlers = {
          'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions?page={page}&pagesize={pagesize}&{filter}',
            {'microservice_id': microservice_id, **data}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table(resp['payload']['data'], self.format, handlers=handlers)
        self.util.print_record_count(resp)

    def start_stop(self, action, microservice, version):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)
        version_data = None

        try:
            version_id = int(version)
        except:
            major, minor, patch = self.util.parse_version_number(version)

            version_data = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'microservice_id': microservice_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']

            if version_data is None or len(version_data) < 1:
                raise click.ClickException(click.style("Version not found", fg='red'))
            else:
                version_id = version_data[0].get('id')
                current_state = version_data[0].get('actualState')

        if version_data is None:
            version_data = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}',\
                {'microservice_id': microservice_id, 'version_id': version_id} ))['payload']['data']

            if version_data is None or len(version_data) < 1:
                raise click.ClickException(click.style("Version not found", fg='red'))

            current_state = version_data.get('actualState')

        if current_state not in ['stopped', 'running', 'error']:
            raise click.ClickException(click.style("To start or stop a version it must have a 'stopped', 'running' or 'error' state. The current state of this version is {}. \
                A version goes through several states after being uploaded to Lumavate. It must finish this process before this command can be run. \
                To check its current state, run 'luma microservice-version ls'.".format(current_state), fg='yellow'))

        if action == 'stop' and current_state != 'running':
            raise click.ClickException(click.style("This version is not running", fg='yellow'))
        if action == 'start' and current_state == 'running':
            raise click.ClickException(click.style("This version is already running", fg='yellow'))

        titles = {
            'id': 'id',
            'actualState': 'Current State',
            'versionNumber': 'Version #',
            'createdAt': 'Created At',
            'createdBy': 'Created By'
        }

        data = {
            "action": action,
            "containerId": microservice_id,
            "id": version_id
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}/{action}',\
            {'microservice_id': microservice_id, 'version_id': version_id, 'action': action}), data=data)

        if self.json:
            click.echo(json.dumps(resp))
            return

        status_id = resp['payload']['data']['results']['statusId']
        progress = 0
        error = None
        if action == 'start':
            bar_name = "Starting microservice"
        else:
            bar_name = "Stopping microservice"

        with click.progressbar(length=100, label=bar_name) as bar:
            while progress < 100 and error is None:
                status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                progress = status_resp['payload']['data']['overallPercent']
                error = status_resp['payload']['data']['errorMessage']
                bar.update(progress)

        click.echo(' ')
        ver_resp = self.util.cli_request('GET',
            self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}', {'microservice_id': microservice_id, 'version_id': version_id}))

        handlers = {
          'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        self.util.print_table([ver_resp['payload']['data']], self.format, handlers=handlers, titles=titles)

    def run(self, microservice, version, command, target):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)
        version_id = self.util.get_version_id(self.object_type, microservice_id, version)

        data = {
            "command": command,
            "target": target
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}/exec',\
            {'microservice_id': microservice_id, 'version_id': version_id}), json=data)

        if self.json:
            click.echo(json.dumps(resp))
            return

        status_id = resp['payload']['data']['statusId']
        progress = 0
        error = None
        with click.progressbar(length=100, label='Executing command') as bar:
            while progress < 100:
                status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                progress = status_resp['payload']['data']['overallPercent']
                bar.update(progress)
                if status_resp['payload']['data']['errorMessage'] is not None:
                    click.echo('There was an error running the command')
                    raise click.ClickException('Status: {}'.format(status_resp['payload']['data']))

        click.echo(' ')
        click.echo('Command Output: ')
        click.echo(' ')
        try:
            status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))['payload']['data']
            if target == 'one':
                summary = status_resp['summary']['results']
                if 'output' in summary[0].keys():
                    if isinstance(summary[0]['output'], list):
                        for x in summary[0]['output']:
                            click.echo(x)
                    else:
                        click.echo(summary)
            else:
                summary = status_resp['summary']['results']
                for x in summary:
                    click.echo(x)

        except:
            click.echo(status_resp)

    def logs(self, microservice, version):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)
        version_id = self.util.get_version_id(self.object_type, microservice_id, version)

        resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}/logs',\
            {'microservice_id': microservice_id, 'version_id': version_id}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        for x in resp['payload']['data']:
            click.echo(x)

    def create(self, microservice, port, microservice_file_path, docker_image,\
        from_version, version_number, version_bump, env, label, format):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)

        if docker_image is not None and docker_image != '':
            if not self.json:
                click.echo("Getting docker image ")
            cu = pyke.ContainerUtil()
            # save_image() returns the file path of the zipped docker image
            microservice_file_path = cu.save_image(docker_image)

        if from_version != '' and from_version is not None:
            try:
                from_version_id = int(from_version)
                version = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{from_version_id}',\
                    {'microservice_id': microservice_id, 'from_version_id': from_version_id} ))['payload']['data']
            except:
                version = self.util.get_version(self.object_type, microservice_id, from_version)

            if version is None:
                raise click.BadParameter(click.style("Microservice version not found", fg='red'), param_hint="--version-number")

            if version_number is not None and version_number != '':
                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

            else:
                version[version_bump] = version[version_bump] + 1
                if version_bump in ['major', 'minor']:
                    version['patch'] = 0
                if version_bump == 'major':
                    version['minor'] = 0

        else:
            version = {}
            if label is None or label == '':
                label = input("Label: ")
                if label is None or label == '':
                    raise click.BadParameter(click.style("If you don't create from an old version then you must provide a --label", fg='red'), param_hint="--label")

            if version_number is None or version_number == '':
                version_number = input("Version Number: ")

                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

                if version_number is None or version_number == '':
                    raise click.BadParameter(click.style("If you don't create from an old version then you must provide a --version-number in the format\
                        '<major: int>.<minor: int>.<patch: int>' ", fg='red'), param_hint="--version-number")

            if microservice_file_path is None or microservice_file_path == '':
                microservice_file_path = input("Widget File Path: ").replace(" ", "")
                if microservice_file_path is None or microservice_file_path == '':
                    raise click.ClickException(click.style("If you don't create from an old version then you must provide --microservice-file-path or --docker-image", fg='red'))

            if port is None or port == '':
                port = int(input("Port: "))
                if port is None or port == '':
                    raise click.ClickException(click.style("If you don't create from an old version then you must provide a --port for your microservice", fg='red'))

            else:
                major, minor, patch = self.util.parse_version_number(version_number)

                version['major'] = major
                version['minor'] = minor
                version['patch'] = patch

        if microservice_file_path is not None and microservice_file_path != '':
            if not self.json:
                click.echo("Uploading image to Lumavate: ")
                click.echo('Image Size: {}'.format(self.util.get_size(microservice_file_path)))

            version['tag'] = None
            version['ephemeralKey'] = self.util.upload_ephemeral(microservice_file_path, 'application/gz')

        version['platformVersion'] = 'v2'
        if 'env' not in version.keys():
            version['env'] = {}

        for var in env:
            try:
                env_dict = json.loads(var)
                version['env'] = { **version['env'], **env_dict}
            except Exception as e:
                print("Could not serialize {}".format(var))
                print(e)

        if label != '' and label is not None:
            version['label'] = label

        if port != '' and port is not None:
            version['port'] = port

        version['instanceCount'] = 1

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions',\
        {'microservice_id': microservice_id}), data=json.dumps(version))

        if self.json:
            click.echo(json.dumps(resp))
            return

        # TO DO: Add progress bar after create for loading/validating microservice
        progress = 0
        status_id = None
        error = None

        if 'results' in resp['payload']['data'].keys():
            status_id = resp['payload']['data']['results']['statusId']
        if status_id:
            with click.progressbar(length=100, label='Creating microservice') as bar:
                while progress < 100 and error is None:
                    status_resp = self.util.cli_request('GET', self.util.build_url('{app}/iot/v1/statuses/{status_id}', {'status_id': status_id}))
                    progress = status_resp['payload']['data']['overallPercent']
                    error = status_resp['payload']['data']['errorMessage']
                    bar.update(progress)


        handlers = {
            'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
        }

        self.util.print_table([resp['payload']['data']], format, handlers=handlers)

        if docker_image is not None and docker_image != '':
            try:
                os.remove(microservice_file_path)
            except:
                click.echo('Failed to pick up our mess...')

    def update(self, microservice, version, label, format):
        microservice_id = self.util.lookup_object_id(self.object_type, microservice)
        version_id = self.util.get_version_id(self.object_type, microservice_id, version)

        post_data = {
          'microserviceId': microservice_id,
          'id': version_id,
          'label': label
        }

        resp = self.util.cli_request('PUT',\
          self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}',
          {'microservice_id': microservice_id, 'version_id': version_id}), data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], format)

    def delete(self, microservice, version, version_mask, format):
        if version is None and version_mask is None:
            version = str(input("Version: "))
            if version is None or version == '':
                raise click.BadParameter(click.style('You must provide either a --version or a --version-mask', fg='red'))

        microservice_id = self.util.lookup_object_id(self.object_type, microservice)

        if version is not None:
            version_id = self.util.get_version_id(self.object_type, microservice_id, version)

            resp = self.util.cli_request('DELETE',\
              self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}', {'microservice_id': microservice_id, 'version_id': version_id}))

            if self.json:
                click.echo(json.dumps(resp))
                return

            self.util.print_table([resp['payload']['data']], format)

        else:
            versions = self.util.get_versions_from_mask(self.object_type, microservice_id, version_mask)

            if self.json:
                resp_list = {}
                resp_list['responses'] = []
                for v in versions:
                    resp = self.util.cli_request('DELETE',
                        self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}', {'microservice_id': microservice_id, 'version_id': v['id']}))
                    resp_list['responses'].append(resp)
                click.echo(json.dumps(resp_list))
                return

            with click.progressbar(versions, label='Deleting Versions') as bar:
                for v in bar:
                    resp = self.util.cli_request('DELETE',
                        self.util.build_url('{app}/iot/v1/containers/{microservice_id}/versions/{version_id}', {'microservice_id': microservice_id, 'version_id': v['id']}))

            self.util.print_table(versions, format)



from collections import Counter
import json
import uuid

USE_HTTP = None

class Status():
  @staticmethod
  def new():
    status = Status()
    return status

  def __init__(self):
    self.summary = None
    self.id = str(uuid.uuid4())[:8]
    self.properties = {}
    self.percent = 0
    self.overall_percent = 0
    self.error_message = None

  def to_json(self):
    return {
      'percent': self.percent,
      'properties': self.properties,
      'errorMessage': self.error_message
    }

  def add_progress(self, percent):
    self.percent += percent
    if self.percent > 100:
      self.percent = 100

  def set_progress(self, new_percent):
    self.percent = new_percent

  def set_property(self, name, value):
    self.properties[name] = value

  def set_error(self, error, message=None, is_complete=True):
    error_text = None
    if error is not None:
      error_text = str(error)

    if message is not None:
      self.error_message = message
      if error_text:
        self.error_message += ': {}'.format(error_text)
    elif error_text is not None:
      self.error_message = error_text

    if is_complete:
      self.set_progress(100)


from pathlib import Path
from pyparsing import *
import dateutil.parser
from cli import pyke
import requests
import platform
import urllib
import click
import copy
import json
import sys
import os


ESC = Literal('\x1b')
integer = Word(nums)
escapeSeq = Combine(ESC + '[' + Optional(delimitedList(integer,';')) +
                oneOf(list(alphas)))

uncolor = lambda s : Suppress(escapeSeq).transformString(s)

class Util:
    def __init__(self, context=None, profile=None):
        self.context = pyke.auth.load_cache()
        if profile is None:
            self.profile = self.context.get('profileName')
        else:
            self.profile = profile

    def login_token(self, renew=False):
        if renew:
            self.context['accessToken'] = self.login()['access_token']
            return self.context['accessToken']

        if self.context is not None:
            return self.context.get('accessToken')

    def login_without_context(self, env, config_data=None):
        if config_data is None:
            config_data = pyke.auth.load_config()
        else:
            config_data = config_data

        env = config_data['envs'][env]

        data = {
          'grant_type': 'client_credentials',
          'audience': env['audience'],
          'client_id': env['clientId'],
          'client_secret': env['clientSecret']
        }

        req_data = pyke.auth.sanatize_data(data)

        token_url = env['token'].replace(" ", "")
        resp = self.cli_request('POST', 'https://' + token_url+ '/oauth/token', json=req_data, external=True)

        if resp.status_code != 200:
            click.echo(json.dumps(env))
            click.echo(' ')

            raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

        return resp.json()['access_token']

    def login(self):
        data = {
          'grant_type': 'client_credentials',
          'audience': self.context['audience'].replace(" ", ""),
          'client_id': self.context['clientId'].replace(" ", ""),
          'client_secret': self.context['clientSecret'].replace(" ", "")
        }

        res = self.cli_request('POST', 'https://{}/oauth/token'.format(self.context['token']), json=data, external=True)

        if res.status_code != 200:
            raise click.ClickException(click.style("Env not configured correctly. Make sure your Client ID and Client Secret are correct.", fg='red'))

        resp = res.json()

        if self.context.get('orgId') is not None:
            url = '{}/auth/v1/session'.format(self.context.get('app'))
            data = {'companyId': self.context.get('orgId')}
            r = requests.post(url, json=data, headers={'Authorization': 'Bearer ' + resp['access_token']})
            if r.status_code != 200:
                raise click.BadParameter(click.style("Not authorized for company id {}".format(self.context.get('orgId')), fg='red'))

        pyke.auth.set_cache_token(resp.get('access_token'))
        pyke.auth.set_profile_token(self.context.get('profileName'), resp.get('access_token'))

        return resp

    def print_record_count(self, data):
        d = data['payload']
        if d['totalItems'] > d['currentItemCount']:
            d['startItem'] = (d['page'] - 1) * d['pageSize'] + 1
            d['endItem'] = d['startItem'] - 1 + d['currentItemCount']
            count_str = 'showing {startItem} - {endItem} of {totalItems} total items (page {page} / {totalPages})'
            click.echo(click.style(count_str.format(**d), fg='yellow'))

    def load_users(self):
        resp = self.cli_request('GET', self.build_url('{app}/auth/v1/users'))

        users = {}
        for u in resp['payload']['data']:
            users[u['id']] = u['emailAddress']

        return users

    def print_table(self, data, format, titles={}, handlers={}):
        headers = {}
        stats = {}
        users = {}

        if 'createdBy' in format or 'lastModifiedBy' in format:
            users = self.load_users()

        for c in data:
            child_records = {}
            for k in c:
                if isinstance(c[k], dict):
                    for sk in c[k]:
                        child_records[k + '__' + sk] = c[k][sk]

            c.update(child_records)

        if len(data) > 0:
            for k in data[0]:
                headers[k] = titles.get(k, k)
        else:
            click.echo('No Records Found')
            return

        for k in headers:
            stats[k] = len(uncolor(headers[k]))

        # TO DO: Cache users and reload when we can't find one
        for c in data:
            for k in c:
                if k in ['createdAt', 'lastModifiedAt']:
                    cd = dateutil.parser.parse(c[k])
                    c[k] = cd.strftime('%x %X')
                if k in ['createdBy', 'lastModifiedBy']:
                    c[k] = users.get(c[k], 'Not Found ({})'.format(c[k]))
                if stats.get(k, 0) < len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k])):
                    stats[k] = len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k]))

        for k in headers:
            headers[k] = titles.get(k, k).ljust(stats[k])

        for c in data:
            for k in c:
                c[k] = str(handlers[k](c[k]) if k in handlers else c[k])
                c[k] = c[k] + ' ' * (stats[k] - len(uncolor(c[k])))

        click.echo(format.format(**headers))

        for c in data:
            print(format.format(**c))

    def check_response(self, resp):
        if resp.status_code == 500:
            raise click.ClickException(str(click.style('Unhandled Exception: ', fg='red')) + str(resp.text))
        elif resp.status_code == 404:
            raise click.ClickException(click.style('Not Found', fg='red'))
        elif resp.status_code == 400:
            message = resp.json()['payload']['error']
            if 'apiField' in resp.json()['payload']:
                message += ' (' + resp.json()['payload']['apiField'] + ')'
            raise click.ClickException(click.style(message, fg='red'))

        return resp

    def get_headers(self):
        return {
            'Authorization': 'Bearer ' + self.login_token(),
            'Content-Type': 'application/json'
        }

    def build_url(self, format, extra_data = {}):
        data = pyke.auth.sanatize_data(self.context)
        env = copy.deepcopy(data)
        env.update(self.context)
        env.update(extra_data)
        return format.format(**env).replace(" ", "+")

    def cli_request(self, method, url, data=None, json=None, headers=None, params=None, external=False):
        req_args = self.build_req_args(data, json, headers, params, external)
        req = self.get_request_func(method)
        url = url.replace(" ", "")
        if data is not None:
            resp = self.check_response(req(url, data=data, **req_args))
        else:
            resp = self.check_response(req(url, **req_args))

        if resp.status_code != 200 and external is False:
            self.login_token(renew=True)
            req_args = self.build_req_args(data, json, headers, params, external)
            resp = self.check_response(req(url, **req_args))
            if resp.status_code != 200:
                raise click.BadParameter(click.style('Invalid Credentials', fg='red'))

        if not external:
            return resp.json()
        else:
            return resp

    def build_req_args(self, data, json_data, headers, params, external=False):
        if external is False:
            req_headers = self.get_headers()
            if headers is not None:
                req_headers.update(headers)
        elif headers is not None:
            req_headers = headers
        else:
            req_headers = None

        req_args = {}
        if json_data is not None:
            req_args['json'] = json_data
        if req_headers is not None:
            req_args['headers'] = req_headers
        if params is not None:
            req_args['params'] = params

        return req_args

    def get_request_func(self, method):
        method = method.lower()
        methods = {
            'get': requests.get,
            'head': requests.head,
            'post': requests.post,
            'put': requests.put,
            'patch': requests.patch,
            'delete': requests.delete
        }

        return methods[method]

    def list_companies(self, env, print_table=True):
        conf = pyke.auth.load_config()

        handlers = {
            'instanceType': lambda x: 'dev' if x == 'cc' else x,
        }

        headers = {
            'Authorization': 'Bearer ' + self.login_without_context(env),
            'Content-Type': 'application/json'
        }

        app_url = conf['envs'][env]['app'].replace(" ", "")
        res = requests.get('{}/auth/v1/me/companies?sort=name'.format(app_url), headers=headers)
        resp = res.json()['payload']['data']

        titles = {
        'id': 'id',
        'name': 'Org Name',
        'instanceType': 'Org Type',
        'isTest': 'Test Org'
        }

        if print_table is True:
            self.print_table(resp, '{id} {name} {instanceType} {isTest}', handlers=handlers, titles=titles)

        return resp

    def get_company_id_from_resp(self, id, resp):
        try:
            company_id = next(int(rec.get('id')) for rec in resp if rec.get('name').lower() == id.lower())
            return company_id
        except:
            raise click.ClickException(click.style("Company not found. You can only share/unshare with companies you have provisioned.", fg='red'))

    def get_company_id(self, name_or_id, resp):
        try:
            company_id = int(name_or_id)
            return company_id
        except:
            return self.get_company_id_from_resp(name_or_id, resp)

    def get_company_name(self, id, resp):
        return next(rec.get('name') for rec in resp if int(rec.get('id')) == id)

    def upload_ephemeral(self, filename, content_type):
        resp = self.cli_request('GET', self.build_url('{app}/iot/v1/files/ephemeral?contentType=' + urllib.parse.quote(content_type)))
        if '/' in filename:
            file_name = filename.rsplit('/', 1)[-1]
        else:
            file_name = filename

        ephemeral_url = resp['payload']['data']['url']
        ephemeral_data = resp['payload']['data']['fields']
        ephemeral_data['key'] = ephemeral_data['key'].replace('${filename}', file_name)
        ephemeral_data['Content-Type'] = content_type

        post_res = requests.post(
            ephemeral_url,
            data=ephemeral_data,
            files={'file': open(filename, 'rb')})

        return ephemeral_data['key']

    def get_current_access_records(self, object, object_id):
        path = '{}s/{}/access'.format(object, object_id)
        resp = self.cli_request('GET', self.build_url('{app}/iot/v1/' + path))
        return resp['payload']['data']

    def lookup_object_id(self, object_type, id, abort_if_null=True):
        if object_type == 'component-set':
            try:
                object_filter = int(id)
                filter_type = 'id'
            except:
                object_filter = id
                filter_type = 'urlRef'

            resp = self.cli_request('GET',
                self.build_url('{app}/iot/v1/component-sets?{filter_type}={object_filter}',
                    extra_data={'filter_type': filter_type, 'object_filter': object_filter}))


            if len(resp.get('payload').get('data', [])) > 0:
                return resp['payload']['data'][0]['id']

            if abort_if_null:
                click.echo(click.style('{} not found'.format(object_type), fg='red'))
                sys.exit()

        else:
            try:
                object_filter = int(id)
                filter_type = 'id'
            except:
                object_filter = id
                filter_type = 'urlRef'

            resp = self.cli_request('GET',
                self.build_url('{app}/iot/v1/containers?{filter_type}={object_filter}',
                    extra_data={'filter_type': filter_type, 'object_filter': object_filter}))

            if len(resp.get('payload').get('data', [])) > 0:
                return resp['payload']['data'][0]['id']

            if abort_if_null:
                click.echo(click.style('{} not found'.format(object_type), fg='red'))
                sys.exit()

    def save_config(self, data):
        with open(self.get_config_file(), 'w+') as outfile:
            json.dump(data, outfile, indent=4, sort_keys=True)

    def get_size(self, path, image=True):
        stats = os.stat(path)
        byte_size = stats.st_size
        if byte_size < 1048576:
            return '{} KB'.format(round(byte_size / 1024, 2))
        elif byte_size < 1073741824:
            return '{} MB'.format(round(byte_size / 1048576, 2))
        else:
            if image is True:
                click.echo('Hey! Thanks for being an awesome Lumavate customer.')
                click.echo('While you are waiting for this image to upload to the platform you should check this out!')
                click.echo('https://hackernoon.com/tips-to-reduce-docker-image-sizes-876095da3b34')
            return '{} GB'.format(round(byte_size / 1073741824, 2))

    def parse_version_number(self, version_number):
        try:
            major, minor, patch = version_number.split('.')
            for x in [major, minor, patch]:
                if x is None:
                    raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))
                if int(x) < 0:
                    raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

            return major, minor, patch
        except:
            raise click.ClickException(click.style("Version numbers must be in the format '<major: int>.<minor: int>.<patch: int>' ", fg='red'))

    def get_config_file(self):
        config_path = str(Path.home()) + '/.luma_cli_config'
        if not os.path.exists(config_path):
            with open(config_path, 'w') as config:
                json.dump({ "envs": {}, "profiles": {} }, config)

        return config_path

    def get_version_id(self, object_type, parent_id, version):
        try:
            version_id = int(version)
        except:
            _version = self.get_version_from_version_number(object_type, parent_id, version)
            if _version is not None:
                version_id = _version.get('id')

        return version_id

    def get_versions_from_mask(self, object_type, parent_id, version):
        if object_type == 'component-set':
            versions = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions',\
                {'parent_id': parent_id}))['payload']['data']
        else:
            versions = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions',\
                {'parent_id': parent_id}))['payload']['data']

        major, minor, patch = version.split('.')

        return sorted([x for x in versions if
            (str(x['major']) == major or major == '*') and
            (str(x['minor']) == minor or minor == '*') and
            (str(x['patch']) == patch or patch == '*')], key=lambda x: (x['major'], x['minor'], x['patch']), reverse=True)

    def get_version_from_version_number(self, object_type, parent_id, version_number):
        major, minor, patch = version_number.split('.')

        if object_type == 'component-set':
            version = self.cli_request('GET', self.build_url('{app}/iot/v1/component-sets/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']
        else:
            version = self.cli_request('GET', self.build_url('{app}/iot/v1/containers/{parent_id}/versions?major={major}&minor={minor}&patch={patch}',\
                { 'parent_id': parent_id, 'major': major, 'minor': minor, 'patch': patch }))['payload']['data']

        if version is not None and len(version) > 0:
            return version[0]

        else:
            raise click.ClickException(click.style('Version Not Found', fg='red'))

    def get_version(self, object_type, parent_id, version):
        versions = self.get_versions_from_mask(object_type, parent_id, version)
        if versions:
            return versions[0]

        return None

    def resolve_default_icon_path(self):
        pwd = os.path.dirname(os.path.realpath(__file__))
        if platform.system() == 'Windows':
            icon_path = pwd.rsplit('\\', 1)[0]
            return '{}\\images\\default-icon.svg'.format(icon_path)
        else:
            icon_path = pwd.rsplit('/', 1)[0]
            return '{}/images/default-icon.svg'.format(icon_path)



from cli import pyke, behavior
import click

@click.group('component-set-version', help="A component-set-version is where the platform looks for components when used in a widget.\
The version also contains other details like CSS includes.")
def component_set_version():
    pass

@click.command('ls')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set')
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {expand__experiences} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def list(component_set, profile, format, json, **kwargs):
    behavior.ComponentSetVersionBehavior(profile=profile, format=format, json=json, **kwargs).list(component_set)

@click.command('components', help="Given a --component-set and a --version, this will return the raw JSON contained in a component-set-version")
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set the version is associated with')
@click.option('--version', '-v', prompt=True)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def components(component_set, profile, version, json):
    behavior.ComponentSetVersionBehavior(profile=profile, json=json).list_components(component_set, version)

@click.command('add', help="While adding a new component-set-version you must always upload a zipped file, even if you use the from-version option")
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set the version is added to')
@click.option('--component-set-file-path', '-path', prompt=True, type=click.Path(), help='Must be the path to a zipped file.')
@click.option('--from-version', '-fv', type=str, help="Take all values from an exsisting version and create a new version. \
Increment the major, minor or patch version by 1 with --major, --minor, --patch flags. Override any old values by passing them in as options. \
When using this option with component-sets you will always need to provide a new, zipped, component set file.")
@click.option('--version', '-v', type=str, default=None, help="The version number of the new version")
@click.option('--patch', 'version_bump', flag_value='patch', default=True)
@click.option('--minor', 'version_bump', flag_value='minor')
@click.option('--major', 'version_bump', flag_value='major')
@click.option('--css-includes', type=str, default=None, multiple=True)
@click.option('--direct-includes', type=str, default=None, multiple=True)
@click.option('--label', '-l', type=click.Choice(['prod', 'dev', 'old']))
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def create(component_set, profile, component_set_file_path, from_version, version,
    version_bump, css_includes, direct_includes, label, format, json):
    behavior.ComponentSetVersionBehavior(profile=profile, format=format, json=json)\
        .create(component_set, component_set_file_path, from_version, version,
            version_bump, css_includes, direct_includes, label)

@click.command()
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set the version is associated with')
@click.option('--version', '-v', prompt=True, help='The ID or the version-number of the version to update')
@click.option('--label', '-l', prompt=True, type=str)
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def update(component_set, version, label, profile, format, json):
    behavior.ComponentSetVersionBehavior(profile=profile, format=format, json=json).update(component_set, version, label)

@click.command('rm')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set the version is associated with')
@click.option('--version-mask', '-vm', help="Delete versions with version numbers that match this version mask")
@click.option('--version', '-v', help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@pyke.auth.profile_context_required
def delete(component_set, version_mask, version, profile, format, json):
    behavior.ComponentSetVersionBehavior(profile=profile, format=format, json=json).delete(component_set, version, version_mask)

component_set_version.add_command(list)
component_set_version.add_command(create)
component_set_version.add_command(delete)
component_set_version.add_command(update)
component_set_version.add_command(components)



from pathlib import Path
from cli import pyke
import requests
import click
import time
import json
import sys
import os

class ComponentSetBehavior:
    def __init__(self, profile=None, component_set=None, format=None, filter=None, page=None, pagesize=None, json=False):
        self.util = pyke.Util(profile=profile)
        if profile is not None:
            self.context = pyke.auth.load_cache()
        self.object_type = 'component-set'
        self.json = json
        self.profile = profile
        self.component_set = component_set
        self.format = format
        self.filter = filter
        self.page = page
        self.pagesize = pagesize


    def list(self):
        data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/iot/v1/component-sets?page={page}&pagesize={pagesize}&{filter}', {**data} ))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table(resp['payload']['data'], self.format)
        self.util.print_record_count(resp)

    def delete(self, component_set):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)

        resp = self.util.cli_request('DELETE',
            self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def create(self, name, url_ref, icon_file):
        if icon_file is None or icon_file == '':
            icon_file = self.util.resolve_default_icon_path()

        post_data = {
          'name': name,
          'urlRef': url_ref,
          'ephemeralKey': self.util.upload_ephemeral(icon_file, 'image/svg+xml')
        }

        resp = self.util.cli_request('POST', self.util.build_url('{app}/iot/v1/component-sets'), data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def update(self, component_set, name, icon_file):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)
        post_data = {}

        if len(name) > 1:
            post_data['name'] = name

        if len(icon_file) > 1:
            post_data['ephemeralKey'] = self.util.upload_ephemeral(icon_file, 'image/svg+xml')

        resp = self.util.cli_request('PUT',
          self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}),
          data=json.dumps(post_data))

        if self.json:
            click.echo(json.dumps(resp))
            return

        self.util.print_table([resp['payload']['data']], self.format)

    def share(self, component_set, add, rm, absolute, current):
        component_set_id = self.util.lookup_object_id(self.object_type, component_set)

        if current:
            if self.json:
                resp = self.util.cli_request('GET',
                    self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access',
                    {'component_set_id': component_set_id}))

                click.echo(resp)
                return
            else:
                resp = self.util.cli_request('GET',
                    self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

                recs = self.util.get_current_access_records('component-set', component_set_id)
                org_names = [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.context.get('orgId'))]
                click.echo(org_names)

                return

        resp = self.util.cli_request('GET',
            self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

        handlers = {
          'sharedWith': lambda x: x.get('granteeId') if x is not None else '',
          'unsharedWith': lambda x: x.get('granteeId') if x is not None else ''
        }

        if absolute:
          absolute_ids = [self.util.get_company_id(x, resp) for x in absolute]
          data = {
            "id": component_set_id,
            "granteeCompanyIds": [x for x in absolute_ids]
          }

          access_resp = self.util.cli_request('PUT', self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access',\
            {'component_set_id': component_set_id}), data=json.dumps(data))

          if self.json:
            click.echo(json.dumps(access_resp))
            return

          recs = self.util.get_current_access_records('component-set', component_set_id)
          access_resp['payload']['data']['resultingGrantees'] =\
            [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.context.get('orgId'))]

          self.util.print_table([access_resp['payload']['data']], self.format)
          return

        recs = self.util.get_current_access_records('component-set', component_set_id)
        sharees = [x.get('granteeCompanyId') for x in recs if x.get('granteeCompanyId') != int(self.context.get('orgId'))]
        if add:
          add_ids = [self.util.get_company_id(x, resp) for x in add]
          sharees.extend(add_ids)

        if rm:
          rm_ids = [self.util.get_company_id(x, resp) for x in rm]
          sharees = [x for x in sharees if x not in rm_ids]

        data = {
          "id": component_set_id,
          "granteeCompanyIds": sharees
        }

        access_resp = self.util.cli_request('PUT',
            self.util.build_url('{app}/iot/v1/component-sets/{component_set_id}/access', {'component_set_id': component_set_id}), data=json.dumps(data))

        if self.json:
          click.echo(json.dumps(access_resp))
          return

        recs = self.util.get_current_access_records('component-set', component_set_id)

        access_resp['payload']['data']['resultingGrantees'] =\
          [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.context.get('orgId'))]

        self.util.print_table([access_resp['payload']['data']], self.format)



from cli import pyke, behavior
import click
import json

@click.group(help='All api subcommands take one url argument. This is the relative path. The CLI will handle your app name and the auth headers. \
For example, to GET all containers (a combination of widgets and microservices) the command would be < luma api get /iot/v1/containers >. \
For PUT and POST commands there is a --data option. The value for --data must be valid json.')
def api():
    pass

@click.command()
@click.argument('url')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@pyke.auth.profile_context_required
def get(url, profile):
    util = pyke.Util(profile=profile)

    resp = util.cli_request('GET',
        util.build_url('{app}{url}', {'url': url}))

    click.echo(json.dumps(resp))

@click.command()
@click.argument('url')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--data', '-d', help='Must be valid json surrounded by single quotes. Will be sent as the payload for the request.')
@pyke.auth.profile_context_required
def put(url, profile, data):
    util = pyke.Util(profile=profile)

    req_data = serialize_data(data)

    resp = util.cli_request('PUT',\
        util.build_url('{app}{url}', {'url': url}), data=json.dumps(req_data))

    click.echo(json.dumps(resp))

@click.command()
@click.argument('url')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@click.option('--data', '-d', help='Must be valid json surrounded by single quotes. Will be sent as the payload for the request.')
@pyke.auth.profile_context_required
def post(url, profile, data):
    util = pyke.Util(profile=profile)

    req_data = serialize_data(data)

    resp = util.cli_request('POST',\
        util.build_url('{app}{url}', {'url': url}), data=json.dumps(req_data))

    click.echo(json.dumps(resp))

@click.command()
@click.argument('url')
@click.option('--profile', '-p', prompt=True, help='The name of the profile to use as the context for the command.', autocompletion=pyke.auth.get_profile_names)
@pyke.auth.profile_context_required
def delete(url, profile):
    util = pyke.Util(profile=profile)

    resp = util.cli_request('DELETE',\
        util.build_url('{app}{url}', {'url': url}))

    click.echo(json.dumps(resp))

api.add_command(get)
api.add_command(put)
api.add_command(post)
api.add_command(delete)

def serialize_data(data):
    req_data = {}
    for var in data:
        try:
            data_dict = json.loads(data)
            req_data = { **req_data, **data_dict}
        except Exception as e:
            click.echo("Could not serialize data")
            click.echo(e)

    return req_data
